//
// Created by andreas on 3/30/18.
//

#include <vector>
#include "ColoredReducer.h"
#include "ColoredPetriNetBuilder.h"

namespace PetriEngine {
    namespace Colored {
        ColoredReducer::ColoredReducer(ColoredPetriNetBuilder *parent)
                : parent(parent)
        {
        }

        void ColoredReducer::reduce(ColoredQueryPlaceAnalysisContext &context, int enableReduction, int timeOut,
                                    bool remove_loops, bool remove_consumers, bool next_safe)
        {
            // TODO: add timeout code
            assert(consistent());
            if (enableReduction == 1) {
                if (!next_safe) {
                    reduceByRuleI(context.getQueryPlaceCount(), context.getQueryTransitionCount(), remove_loops, remove_consumers);
                }
            }
        }

        bool ColoredReducer::reduceByRuleI(uint32_t *placeInQuery, uint32_t* transitionInQuery, bool remove_loops, bool remove_consumers) {
            bool reduced = false;
            if (remove_loops) {
                std::vector<uint32_t> wtrans;
                std::vector<bool> tseen(parent->getTransitionCount(), false);

                for (uint32_t t = 0; t < parent->getTransitionCount(); ++t) {
                    if (hasTimedOut()) return false;
                    if (transitionInQuery[t] > 0) {
                        wtrans.push_back(t);
                        tseen[t] = true;
                    }
                }

                for (uint32_t p = 0; p < parent->getPlaceCount(); ++p) {
                    if(hasTimedOut()) return false;
                    if(placeInQuery[p] > 0)
                    {
                        const Place& place = parent->_places[p];
                        for(auto t : place.consumers)
                        {
                            if(!tseen[t])
                            {
                                wtrans.push_back(t);
                                tseen[t] = true;
                            }
                        }
                        for(auto t : place.producers)
                        {
                            if(!tseen[t])
                            {
                                wtrans.push_back(t);
                                tseen[t] = true;
                            }
                        }
                    }
                }

                std::vector<bool> pseen(parent->getPlaceCount(), false);
                while(!wtrans.empty())
                {
                    if(hasTimedOut()) return false;
                    auto t = wtrans.back();
                    wtrans.pop_back();
                    const Transition& trans = parent->_transitions[t];
                    for(auto a : trans.pre)
                    {
                        auto& arc = parent->_arcs[a];
                        const Place& place = parent->_places[arc.place];
                        for(auto pt : place.producers)
                        {
                            if(!tseen[pt])
                            {
                                Transition& trans = parent->_transitions[pt];
                                auto it = trans.pre.begin();
                                for(; it != trans.pre.end(); ++it)
                                    if(parent->_arcs[*it].place >= arc.place) break;

                                auto& i = parent->_arcs[*it];
                                if(it != trans.pre.end() && i.place == arc.place)
                                {
                                    auto it2 = trans.post.begin();
                                    for(; it2 != trans.post.end(); ++it2)
                                        if(parent->_arcs[*it2].place >= arc.place) break;
                                    auto ap = i.expr->approximate();
                                    auto ap2 = parent->_arcs[*it2].expr->approximate();
                                    if(!ap.empty() && !ap2.empty() && ap >= ap2) continue;
                                }

                                tseen[pt] = true;
                                wtrans.push_back(pt);
                            }
                        }

                        for(auto pt : place.consumers)
                        {
                            if(!tseen[pt] && (!remove_consumers || placeInQuery[pt] > 0))
                            {
                                tseen[pt] = true;
                                wtrans.push_back(pt);
                            }
                        }
                        pseen[arc.place] = true;
                    }
                }

                for(size_t t = 0; t < parent->getTransitionCount(); ++t)
                {
                    if(!tseen[t] && !parent->_transitions[t].skip)
                    {
                        skipTransition(t);
                        reduced = true;
                    }
                }

                for(size_t p = 0; p < parent->getPlaceCount(); ++p)
                {
                    if(!pseen[p] && !parent->_places[p].skip && placeInQuery[p] == 0)
                    {
                        assert(placeInQuery[p] == 0);
                        skipPlace(p);
                        reduced = true;
                    }
                }

                if(reduced)
                    ++_ruleI;
            } else {
                for (uint32_t p = 0; p < parent->getPlaceCount(); ++p) {
                    if (hasTimedOut()) return false;
                    const Place &place = parent->_places[p];
                    if (place.skip) continue;
                    if (placeInQuery[p] > 0) continue;
                    if (!place.consumers.empty()) continue;

                    ++_ruleI;
                    reduced = true;

                    std::vector<uint32_t> toRem;
                    for (auto t : place.producers) {
                        auto &trans = parent->_transitions[t];
                        if (trans.post.size() > 1) continue;
                        bool ok = true;
                        for (auto &a : trans.pre) {
                            auto &arc = parent->_arcs[a];
                            if (placeInQuery[arc.place] > 0) {
                                ok = false;
                            }
                        }
                        if (ok) toRem.push_back(t);
                    }
                    skipPlace(p);

                    for (auto t : toRem) {
                        skipTransition(t);
                    }
                    assert(consistent());
                }
            }

            return reduced;
        }

        Transition &ColoredReducer::getTransition(uint32_t transition) {
            return parent->_transitions[transition];
        }

        std::vector<uint32_t>::iterator ColoredReducer::getOutArc(Transition& t, uint32_t place) {
            return t.getOutArc(parent->_arcs, place);
        }

        std::vector<uint32_t>::iterator ColoredReducer::getInArc(uint32_t place, Transition& t) {
            return t.getInArc(parent->_arcs, place);
        }

        void ColoredReducer::eraseTransition(std::vector<uint32_t> &set, uint32_t el) {
            auto lb = std::lower_bound(set.begin(), set.end(), el);
            assert(lb != set.end());
            assert(*lb == el);
            set.erase(lb);
        }

        void ColoredReducer::skipPlace(uint32_t place) {
            ++_removedPlaces;
            Place& pl = parent->_places[place];
            assert(!pl.skip);
            pl.skip = true;
            for(auto& t : pl.consumers) {
                Transition& trans = getTransition(t);
                auto ait = getInArc(place, trans);
                if(ait != trans.pre.end() && parent->_arcs[*ait].place == place) {
                    parent->_arcs[*ait].skip = true;
                    trans.pre.erase(ait);
                }
            }

            for(auto& t : pl.producers) {
                Transition& trans = getTransition(t);
                auto ait = getOutArc(trans, place);
                if(ait != trans.post.end() && parent->_arcs[*ait].place == place) {
                    parent->_arcs[*ait].skip = true;
                    trans.post.erase(ait);
                }
            }
            pl.consumers.clear();
            pl.producers.clear();
            assert(consistent());
        }

        void ColoredReducer::skipTransition(uint32_t t) {
            ++_removedTransitions;
            Transition& trans = getTransition(t);
            assert(!trans.skip);
            for(auto p : trans.post) {
                auto& arc = parent->_arcs[p];
                arc.skip = true;
                eraseTransition(parent->_places[arc.place].producers, t);
            }
            for(auto p : trans.pre) {
                auto& arc = parent->_arcs[p];
                arc.skip = true;
                eraseTransition(parent->_places[arc.place].consumers, t);
            }
            trans.post.clear();
            trans.pre.clear();
            trans.skip = true;
            assert(consistent());
        }

        bool ColoredReducer::consistent() {
#ifndef NDEBUG
            size_t strans = 0;
            for(size_t i = 0; i < parent->getTransitionCount(); ++i) {
                Transition& t = parent->_transitions[i];
                if(t.skip) ++strans;
                assert(t.is_sorted(parent->_arcs));
                assert(!t.skip || (t.pre.size() == 0 && t.post.size() == 0));
                for(auto a : t.pre) {
                    auto& arc = parent->_arcs[a];
                    assert(arc.expr->weight() > 0);
                    Place& p = parent->_places[arc.place];
                    assert(!p.skip);
                    assert(std::find(p.consumers.begin(), p.consumers.end(), i) != p.consumers.end());
                }
                for(auto a : t.post)
                {
                    auto& arc = parent->_arcs[a];
                    assert(arc.expr->weight() > 0);
                    Place& p = parent->_places[arc.place];
                    assert(!p.skip);
                    assert(std::find(p.producers.begin(), p.producers.end(), i) != p.producers.end());
                }
            }

            assert(strans == _removedTransitions);

            size_t splaces = 0;
            for(size_t i = 0; i < parent->getPlaceCount(); ++i) {
                Place& p = parent->_places[i];
                if(p.skip) ++splaces;
                assert(std::is_sorted(p.consumers.begin(), p.consumers.end()));
                assert(std::is_sorted(p.producers.begin(), p.producers.end()));
                assert(!p.skip || (p.consumers.size() == 0 && p.producers.size() == 0));

                for(uint c : p.consumers) {
                    Transition& t = parent->_transitions[c];
                    assert(!t.skip);
                    auto a = getInArc(i, t);
                    assert(a != t.pre.end());
                    auto& arc = parent->_arcs[*a];
                    assert(arc.place == i);
                }

                for(uint prod : p.producers) {
                    Transition& t = parent->_transitions[prod];
                    assert(!t.skip);
                    auto a = getOutArc(t, i);
                    assert(a != t.post.end());
                    auto& arc = parent->_arcs[*a];
                    assert(arc.place == i);
                }
            }
            assert(splaces == _removedPlaces);
#endif
            return true;
        }

        ColoredQueryPlaceAnalysisContext::ColoredQueryPlaceAnalysisContext(
                const std::unordered_map<std::string, uint32_t> &places,
                const std::unordered_map<std::string, uint32_t> &tnames, const PetriNet *net,
                const std::unordered_map<std::string, std::unordered_map<uint32_t, std::string>> &cplaces,
                const std::unordered_map<std::string, std::vector<std::string>> &ctnames, bool colored)
                : ColoredAnalysisContext(places, tnames, net, cplaces, ctnames, colored, false)
        {
            _placeInQuery.resize(places.size(), 0);
            _transitionInQuery.resize(tnames.size(), 0);
        }

        bool ColoredQueryPlaceAnalysisContext::resolvePlace(const std::string &place,
                                                            std::unordered_map<uint32_t, std::string> &out) {
            auto name = _placeNames.find(place);
            if (name != _placeNames.end()) {
                ++_placeInQuery[name->second];
                return true;
            }
            return false;
        }

        bool ColoredQueryPlaceAnalysisContext::resolveTransition(const std::string &transition,
                                                                 std::vector<std::string> &out) {
            auto name = _transitionNames.find(transition);
            if (name != _transitionNames.end()) {
                ++_transitionInQuery[name->second];
                return true;
            }
            return false;
        }

    }
}